
/*Listing 13-5.  C Portion of Package, SERIAL.C*/
/*---------------------------------------------------------------------*/
/*
 * Filename:        S E R I A L . C
 * Purpose:         An asynchronous communication package for
 *                  IBM PC and compatibles.  Supports asynchronous
 *                  communication adapters based on Intel 8250
 *                  UART (or compatible).
 * Language:        Microsoft C 4.0
 * Usage:           This is a set of routines that provides basic
 *                  functions for asynchronous communication.
 *                  Call s_setvals to set parameters and s_getvals
 *                  see parameters values.  Call s_setup to setup
 *                  port for communications.  Use s_sendchar and
 *                  s_rcvchar to send and receive characters
 *                  respectively.  A BREAK may be sent by calling
 *                  s_sendbreak.
 *
 * Copyright  1986, 1987 Nabajyoti Barkakati
 *                         Silver Sprint, MD 20904
 * Right to use, copy, and modify this code is granted for personal
 * noncommercial use, provided that this copyright disclosure is   
 * retained on all copies.  All other rights reserved.
 *
 * Data Started:        10-DEC-1986
 * Revisions:  V1.00     9-FEB-1987--First working version. (NB)
*/
/*-------------------------------------------------------------------*/
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <dos.h>
#include <conio.h>

#define TRUE 1
#define FALSE 0
#define EOS '\0'

/* Definitions for the 8259 Programmable Interrupt Controller */
#define P8259_0    0x20     /* Address of int control register */
#define P8259_1    0x21     /* Address of int mask register */
#define END_OF_INT 0x20     /* Nonspecific EOI              */

/* Define some ASCII characters */
#define NUL_ASCII   (0)
#define XON_ASCII   (0x11)
#define XOFF_ASCII  (0x13)
#define DEL_ASCII   (0x7f)

/* Address of BIOS data area (segment 40h, ofset 0) */
#define BIOS_DATA   ((int far *)(0x400000L))

/* Some interrupt vectors */
#define BIOS_RS232  0x14
#define BIOS_TOD    0x1a

/* Defines for BREAK */
#define BREAK_ON  0x40
#define BREAK_OFF 0xbf

/* Time of day interrupts occur once every 18.2 seconds */
#define TOD_INTERVAL (1000.0/18.2) /* Interval in milliseconds */

/* The address of the comm port is in the integer
 * 'comport'.  This variable is initialized by reading from 
 * the BIOS data area at segment 0x40.
 */
#define IER                (comport + 1)  /* Interrupt Enable Register */
#define IIR                (comport + 2)  /* Interrupt Identification  */
#define LCR                (comport + 3)  /* Line Control Register     */
#define MCR                (comport + 4)  /* Modem Control Register    */
#define LSR                (comport + 5)  /* Line Status Register      */
#define MSR                (comport + 6)  /* Modem Status Register     */

/* Individual Interrupt Enable Numbers */
#define RDAINT  1
#define THREINT 2
#define RLSINT  4
#define MSINT   8

/* Moden Control Register value */
#define MCRALL 15         /* (DTR, RTS, OUT1 and OUT2 = 1)  */
#define MCROFF 0          /* Everything off                 */

/* Interrupt Enable Register value to turn on/off all int.  */
#define IERALL  (RDAINT+THREINT+RLSINT+MSINT)
#define IEROFF  0

/* Some masks for turning interrupts off */
#define THREOFF 0xfd

/* Interrupt identification numbers */
#define MDMSTATUS   0
#define TXREGEMPTY  2
#define RXDATAREADY 4
#define RLINESTATUS 6

/* Some flags */
#define XON_RCVD   1
#define XOFF_RCVD  0
#define XON_SENT   1
#define XOFF_SENT  0

/* Hi and low percentages for xon-xoff trigger */
#define HI_TRIGGER(x)  (3*x/4)
#define LO_TRIGGER(x)  (x/4)
/* Function to get bit 0 of an integer */
#define bit0(i)      (i & 0x0001)

/* Function to turn on interrupt whose "Interrupt Enable Number:
 * is 'i', in case it has been disabled.  For example, the 
 * THRE interrupt is disable when an XOFF is received from the 
 * remote system.
 */
#define turnon_int(i,j)  if(((j=inp(IER))&i)==0)outp(IER,(j|i))

#define report_error(s)  fprintf(stderr,s)

typedef struct QTYPE {
    int count;
    int front;
    int rear;
    int elemsize;
    int maxsize;
    char *data;
} QTYPE;

typedef struct VTBLTYPE {
    char *vname;
    int value;
} VTBLTYPE;

typedef struct NTBLTYPE {
    char     *pname;
	VTBLTYPE *vtblentry;
    int      *p_value;
    VTBLTYPE *p_vtable;
    int      vtblsize;
} NTBLTYPE;

/* Functions accessible by user */
int  s_sendchar(), s_rcvchar(), s_setvals(), s_setup(),
     s_cleanup(), s_txqempty();
void s_sendbreak(), s_delay();
char *s_getvals();

/*Global status indicators */
int  s_linestatus, s_modemstatus, s_rbocount, s_totalcount,
     s_rlscount, s_rdacount, s_trmtycount, s_mscount; 

extern QTYPE *q_setup();
extern char *q_getfrom();
extern int q_puton(), q_reset();

/* Use interrupt attribute to define C function as interrupt handler */
void interrupt cdecl far s_mainhndlr();
void far *oldvect_ptr;

static void s_rls(), s_rda(), s_trmty(), s_ms();
static QTYPE *txq, *rxq;
static union REGS xr, yr;
static struct SREGS sr;

static VTBLTYPE b_table[] = {
     "110", 0,
     "150", 1,
     "300", 2,
     "600", 3,
    "1200", 4,
    "2400", 5,
    "4800", 6,
    "9600", 7,
};

static VTBLTYPE p_table[] = {

     "none", 0,
     "odd",  1,
     "even", 3
};

static VTBLTYPE s_table[] = {
     "1", 0,
     "2", 1
};
static VTBLTYPE w_table[] = {
     "7", 2,
     "8", 3
};

static VTBLTYPE port_table[] = {
     "1", 0,
     "2", 1
};

static VTBLTYPE onoff_table[] = {
     "off", 0,
     "on",  1
};

static int   port_number = 0,
             comport=0,
             baudrate = 4,
             parity = 0,
             stopbits = 0,
             wordlength = 3,
             txqsize = 1024,
             rxqsize = 2048,
             enable_xonxoff = 1,
             rcvd_xonxoff = XON_RCVD,
             sent_xonxoff = XON_SENT,
             null_ignore = TRUE,
             del_ignore = TRUE,
             bit7_1_rx = FALSE,
             bit7_0_rx = TRUE,
             bit7_1_tx = FALSE,
			 bit7_0_tx = FALSE,
			 send_xon = FALSE,
             send_xoff = FALSE,
             pckd_comparams = 0x83,
             int_number = 12;
             int_enable_mask = 0xef,
             int_disable_mask = 0x10;

static NTBLTYPE pname_table[] = {
	"baud",         &(b_table[4]),      &baudrate,
			b_table, (sizeof(b_table)/sizeof(VTBLTYPE)),
	"parity",       &(p_table[0]),      &parity,
			p_table, (sizeof(p_table)/sizeof(VTBLTYPE)),
	"stopbits",     &(s_table[0]),  &stopbits,
			s_table, (sizeof(s_table)/sizeof(VTBLTYPE)),
	"wordlength",   &(w_table[1]),  &wordlength,
			w_table, (sizeof(w_table)/sizeof(VTBLTYPE)),
	"port",         &(port_table[0]), &port_number,
			port_table, (sizeof(port_table)/sizeof(VTBLTYPE)),
	"xonoff",       &(onoff_table[1]), &enable_xonxoff,
			onoff_table, (sizeof(onoff_table)/sizeof(VTBLTYPE)),
	"null_ignore",  &(onoff_table[1]), &null_ignore,
			onoff_table, (sizeof(onoff_table)/sizeof(VTBLTYPE)),
	"del_ignore",   &(onoff_table[1]), &del_ignore,
			onoff_table, (sizeof(onoff_table)/sizeof(VTBLTYPE)),
	"8thbit0_on_rx", &(onoff_table[1]), &bit7_0_rx,
			onoff_table, (sizeof(onoff_table)/sizeof(VTBLTYPE)),
	"8thbit1_on_rx", &(onoff_table[0]), &bit7_1_rx,
			onoff_table, (sizeof(onoff_table)/sizeof(VTBLTYPE)),
	"8thbit0_on_tx", &(onoff_table[0]), &bit7_0_tx,
			onoff_table, (sizeof(onoff_table)/sizeof(VTBLTYPE)),
	"8thbit1_on_tx", &(onoff_table[0]), &bit7_1_tx,
			onoff_table, (sizeof(onoff_table)/sizeof(VTBLTYPE)),
};

#define  pnmtbl_size  (sizeof(pname_table)/sizeof(NTBLTYPE))
/*-------------------------------------------------------------*/
/* s _ m a i n d l r
 * Main interrupt handler for all serial port interrupts.
 */
void interrupt cdecl far s_mainhndlr()
{
int c;
    register int int_id, intmask;
    while (TRUE) {
/* Read the interrupt identification register, IIR */
       int_id = inp(IIR);
       if (bit0(int_id) == 1) {
/* If bit 0 is 1, then no interrupts pending.  Send an end of
 * interrupt signal to the 8259A Programmable Interrupt
 * Controller and then return.
 */
             outp(P8259_0, END_OF_INT);
             return;
}
s_totalcount++;
if (int_id >= RXDATAREADY) turnon_int(THREINT, intmask);

/* Process interrupt according to id.
 * The following list is in increasing order of priority.
 */
         switch (int_id) {
             case MDMSTATUS:     s_ms();
                                 break;
             case TXREGEMPTY:    s_trmty();
                                 break;
		 case RXDATAREADY:  s_rda();
                                 break;
		 case RLINESTATUS:   s_rls();
                                 break;
/* Just fall through if id is none of the above */
        }
    }
}
/*------------------------------------------------------------*/
/* s_ r l s
 * Process a "receive line status" interrupt
 */
static void s_rls()
{
    register int intmask;
/* Read line status and save it in 's_linestatus' */
    s_linestatus = inp(LSR);
    s_rlscount++;
}
/*-----------------------------------------------------------*/
/* s _ r d a 
 * Process a "receive data available" interrupt
 */
static void s_rda()
{
    register int intmask;
    char c;
 /* Read from comport */
     c = inp(comport);
     s_rdacount++;
	 if(enable_xonxoff) {
         if(c == XON_ASCII) {
			  rcvd_xonxoff = XON_RCVD;
/* Turn on THRE interrupt if it's off. */
            turnon_int(THREINT, intmask);
            return;
        }
        if(c == XOFF_ASCII) {
			rcvd_xonxoff = XOFF_RCVD;
/* Turn off THRE interrupts. */
            intmask = inp(IER);
            if (intmask & THREINT) outp(IER, intmask & THREOFF);
            return;
        }
   }
   if(null_ignore && c == NUL_ASCII) return;
   if(del_ignore && c == DEL_ASCII) return;
   if(bit7_1_rx) c |= 0x80;
   if(bit7_0_rx) c &= 0x7f;
   if(q_puton(rxq, &c) == 0){
/* Increment receive buffer overflow count */
       s_rbocount++;
   }
/* Check if queue is almost (75%) full */
    if(enable_xonxoff){
        if(rxq->count >= HI_TRIGGER(rxqsize) &&
			sent_xonxoff != XOFF_SENT ) {
/* Set flag to send XOFF */
            send_xoff = TRUE;
/* and turn on THRE interrupts so that we can send the XOFF */
            turnon_int(THREINT, intmask);
            }
    }
}
/*---------------------------------------------------------------*/
/* s _ t r m t y 
 * Process a "transmit holding register empty" interrupt
 */
static void s_trmty()
{
    char c;
    register int ierval;

    s_trmtycount++;

    if (send_xoff == TRUE) {
        outp(comport, XOFF_ASCII);
        send_xoff = FALSE;
        sent_xonxoff = XOFF_SENT;
        return;
    }
    if (send_xon == TRUE) {
        outp(comport, XON_ASCII);
        send_xon = FALSE;
        sent_xonxoff = XON_SENT;
        return;
    }

/* Put a character into the transmit holding register */
    if( q_getfrom(txq, &c) != NULL){
        if(bit7_1_tx) c |= 0x80;
        if(bit7_0_tx) c &= 0x7f;
        outp(comport, c);
        return;
    }
/* Nothing to send--turn off THRE interrupts */
    ierval = inp(IER);
    if (ierval & THREINT) outp(IER, ierval & THREOFF);
}
/*-----------------------------------------------------------------*/
/* s _ m s
 * Process a "modem status" interrupt
 */
static void s_ms()
{
/* Read moden status and save in 's_modemstatus' */
    s_modemstatus = inp(MSR);
    s_mscount++;
}
/*----------------------------------------------------------------*/
/* s _ s e t u p
 * Sets up everything for communication.  Call this routine
 * after parameter values have been specified (by s_setparams).
 * Return 1 if setup sucessful, else return 0.
 */
int s_setup()
{
int i;
    static void s_intinit();

    if (port_number < 0 || port_number > 1)
                 report_error("Invalid port number!\n");
/* Get port address from BIOS data area and save in 'comport' */
	comport = *(BIOS_DATA + port_number);
    if (comport == 0){
        report_error("BIOS could not find port!\n");
        return(0);
    }
/* Set up masks for 8259A PIC.  To enable interrupt from the 
 * port this mask in ANDed with the mask register at 21h.
 * To disable, OR the disable mask with the mask register.
 * The interrupt number is 8+the IRQ level of the interrupt.
 * Com port 1 has IRQ 4, port 2 has IRQ 3.
 */
    if (port_number == 0) {
        int_enable_mask = 0xef;
        int_disable_mask = 0x10;
        int_number = 12;
    }
    if (port_number == 1) {
	int_enable_mask = 0xf7;
        int_disable_mask = 8;
        int_number = 11;
    }
/* Get old interrupt vector and save it.  Use DOS function 35h */
oldvect_ptr = _dos_getvect(int_number);
    
/* Install new interrupt service routine, use DOS function 25h */

_dos_setvect(int_number, s_mainhndlr);

/* Set up communication parameters */
    s_setcommparams();

/* Setup the transmit and receive queues */
    if ( (txq = q_setup(txqsize, sizeof(char))) == NULL   ){
        report_error("Error creating transmit queue!\n");
        return(0);
    }
    if ( (rxq = q_setup(rxqsize, sizeof(char))) == NULL   ){
        report_error("Error creating receive queue!\n");
        return(0);
    }

/* Reset all counts and flags. */
    s_rbocount = 0;
    s_totalcount = 0;
    s_rlscount = 0;
    s_rdacount = 0;
    s_trmtycount = 0;
    s_mscount = 0;

    rcvd_xonxoff = XON_RCVD;
	if (sent_xonxoff == XOFF_SENT) {
		send_xon = TRUE;
    }
    else {
        send_xon = FALSE;
    }
    send_xoff = FALSE;

/* Turn on interrupts from the comm port and setup 8259A */
_disable();
    s_intinit();
_enable();

    return(1);
}
/*-------------------------------------------------------------------*/
/* s _ i n t i n i t
 * Start up interrupts from the serial board.  Also set up
 * 8259A to accept the interrupts.  This routine should be
 * invoked with the interrupts turned off, i.e., inside a
 * (CLI---STI) pair.
 */
static void s_intinit() 
{
    int intmask;

/* Set up moden control register (port = MCR) */
    outp(MCR, MCRALL);

/* Enable all interrupts on the serial card (port = IER) */
    outp(IER, IERALL);

/* Read 8259A's interrupt mask register and write it back after
 * ANDing with int_enable_mask.
 */
    intmask = inp(P8259_1) & int_enable_mask;
    outp(P8259_1, intmask);
}
/*------------------------------------------------------------------*/
/* s _ c l e a n u p
 * Cleanup after comm session is done.  Turns off all interrupts.
 */
 int s_cleanup()
{
     static void s_intoff();

/* Turn off interrupts from serial card */
_disable();
s_intoff();
_enable();

/*  Restore original interrupt vectors */
_dos_setvect(int_number, oldvect_ptr);

/* Deallocate space used by the transmit and receive queues */
    q_reset(txq);
    q_reset(rxq);
}
/*----------------------------------------------------------------*/
/* s _ i n t o f f
 * Turn off all interrupts after the comm session is done.
 * Should be called with interrupts cleared.
 */
static void s_intoff()
{
    int intmask;
  
/* First reset the Interrupt Enable Register on the comm card */
    outp(IER, IEROFF);

/* Turn off all bits of Modem Control Register */
    outp(MCR, MCROFF);

/* Next disable 8259A from recognizing ints at this IRQ level  */
    intmask = inp(P8259_1) | int_disable_mask;
    outp(P8259_1, intmask);
}
/*-------------------------------------------------------------*/
/* s _ s e t c o m m p a r a m s
 * Set up basic communication parameters by using BIOS interrupt    
 * number 14h, function 0 (ah=0).
 */
int s_setcommparams ()
{
/* Set up communication port parameters.  Use BIOS INT 14h, AH=0 */
   pckd_comparams = (baudrate << 5) | (parity << 3) |
                  (stopbits << 2) | (wordlength);
  xr.h.ah = 0;
  xr.h.al = pckd_comparams;
  xr.x.dx = port_number;
  int86(BIOS_RS232, &xr, &yr);

/* Copy status into "s_linestatus" and "s_modemstatus" */
    s_linestatus = yr.h.ah;
    s_modemstatus = yr.h.al;
}
/*------------------------------------------------------------*/
/* s _ s e t v a l s
 * Set the value string for a named parameter.
 * Return 1 if ok, 0 otherwise.
 */
int s_setvals(name, vals)
char *name, *vals;
{
     static  int  s_vtblsrch();
     int i, j, retval;

     retval = 0;
/* First search table for parameter name */
    for (i=0; i<pnmtbl_size; i++) {
		if (strcmpi(pname_table[i].pname, name) == 0) {
/* Now search the table for the value string */
    if ((j  = s_vtblsrch(vals, pname_table[i].p_vtable,
					 pname_table[i].vtblsize)) != -1) {
       *(pname_table[i].p_value) =
           (pname_table[i].p_vtable)[j].value;
       pname_table[i].vtblentry = 
          &((pname_table[i].p_vtable)[j]);
       retval = 1;
    }
    break;
  }
}
return(retval);
}
/*------------------------------------------------------------*/
/* s _ v t b l s r c h
 * Searches a value string in specified table.  Returns index if
 * found.  Returns -1 if string not in table. (Internal Routine)
 */
static int s_vtblsrch(string, table, size)
char *string;
VTBLTYPE *table;
int size;
{
	int i;
	for (i = 0; i < size; i++) {
		if ( (strcmpi(string, table[i].vname)) == 0) {
			return(i);
		}
	}
	return(-1);
}

/*---------------------------------------------------------*/
/* s _ g e t v a l s 
 * Return the value string corresponding to a named parameter.
 * Returns NULL in case of error.
 */
char *s_getvals(name)
char *name;
{
    int i;
    for (i=0; i<pnmtbl_size; i++) {
        if (strcmpi(pname_table[i].pname, name) == 0) {
           return((pname_table[i].vtblentry)->vname);
        }
    }
    return(NULL);
}
/*---------------------------------------------------------*/
/* s _ s e n d c h a r
 * Puts a character into transmit queue.  Returns 1 if all's ok,
 * o if there were problems.
 */
#pragma check_stack(off)
int s_sendchar(ch)
int ch;
{
   int retval, intmask;

   retval = q_puton(txq, &ch);
/* Turn on THRE interrupt if it's off and an XOFF was not received */
	if (rcvd_xonxoff != XOFF_RCVD) turnon_int(THREINT,intmask);
    return(retval);
}
#pragma check_stack(on)
/*---------------------------------------------------------*/                  
/* s _ r c v c h a r 
 * Returns a character from the receive queue.
 * Returns -1 of queue is empty.
 */
int s_rcvchar()
{
    int ch, intmask;
/* If we had sent an XOFF earlier, we might have to send an XON */
     if(enable_xonxoff) {
         if(rxq->count <= LO_TRIGGER(rxqsize) &&
			sent_xonxoff != XON_SENT ) {
              send_xon = TRUE;
              turnon_int(THREINT,intmask);
              printf("int on\n");
     }
   }
   if ( q_getfrom(rxq, &ch) == NULL) {
      return(-1);
   }
   else {
       return(ch);
   }
}
/*---------------------------------------------------------*/
/* s _ t x q e m p t y
 * Returns 1 if txq (transmit queue) is empty, else returns 0.
 * Need when sending file.
 */
int s_txqempty()
{
    if (txq->count == 0) {
       return (TRUE);
    }
    else {
         return (FALSE);
    }
}
/*----------------------------------------------------------*/
/* s _ s e n d b r e a k 
 * Send a break signal (hold line in spacing state for 250 ms).
 */
void s_sendbreak()
{
     int lcrval;
/* Turn on bit 6 of LCR to initiate BREAK signal */
	lcrval = inp(LCR) | BREAK_ON;
    outp(LCR, lcrval);
    s_delay(250.0);
/* Turn off bit 6 of LCR to end the BREAK signal */
    lcrval = inp(LCR) & BREAK_OFF;
    outp(LCR, lcrval);
}
/*----------------------------------------------------------*/
/* s _ d e l a y
 * Delay for 'd_ms" milliseconds.  Resolution ~55 millseconds.
 * WARNING:  This is a very "coarse" timer.
 */
void s_delay(d_ms)
double d_ms;
{
    int ticks;
    long oldcount, newcount;
/* Get timer ticks for this delay */
    ticks = (int)(d_ms / TOD_INTERVAL + 0.5);
    xr.h.ah = 0;
    int86(BIOS_TOD, &xr, &yr);
/* On return CX has high count and DX has low count.
 * The high count is incremented every 65, 536 low counts.
 */
   oldcount = (long)(( (long)(yr.x.cx) << 16) |
             (unsigned) yr.x.dx);
   newcount = oldcount;
/* Now keep checking count until difference between new and 
 * old counts is 'ticks'.
 */
    while ((newcount - oldcount) < ticks) {
	   int86(BIOS_TOD, &xr, &yr);
       if (yr.h.al !=0) {
/* The timer has crossed 24 hours */
            newcount = (long)((24L << 16) |  (unsigned) yr.x.dx);
       }
	  else {
          newcount = (long)(( (long)(yr.x.cx) << 16) |
                     (unsigned) yr.x.dx);
      }
           }
}
/*-------------------------END OF FILE:  SERIAL.C-----------------------*/
